@@ -21,7 +21,7 @@ struct http_connection_info {
 static void http_send_header(struct connection *c);
 static void http_get_header(struct connection *c);
 static void test_restart(struct connection *c);
-static void add_user_agent(unsigned char **hdr, int *l);
+static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url);
 static void add_referer(unsigned char **hdr, int *l, unsigned char *url, unsigned char *prev_url);
 static void add_accept(unsigned char **hdr, int *l);
 static void add_accept_language(unsigned char **hdr, int *l, struct http_connection_info *info);
@@ -116,6 +116,164 @@ int get_http_code(unsigned char *head, int *code, int *version)
 	return 0;
 }
 
+static const unsigned char *months[12] = {
+	cast_uchar "Jan",
+	cast_uchar "Feb",
+	cast_uchar "Mar",
+	cast_uchar "Apr",
+	cast_uchar "May",
+	cast_uchar "Jun",
+	cast_uchar "Jul",
+	cast_uchar "Aug",
+	cast_uchar "Sep",
+	cast_uchar "Oct",
+	cast_uchar "Nov",
+	cast_uchar "Dec"
+};
+
+#ifdef HAVE_GMTIME
+static const unsigned char *days[7] = {
+	cast_uchar "Sun",
+	cast_uchar "Mon",
+	cast_uchar "Tue",
+	cast_uchar "Wed",
+	cast_uchar "Thu",
+	cast_uchar "Fri",
+	cast_uchar "Sat",
+};
+#endif
+
+time_t parse_http_date(unsigned char *date)	/* this functions is bad !!! */
+{
+#ifdef HAVE_MKTIME
+	time_t t = 0;
+	time_t offset;
+	/* Mon, 03 Jan 2000 21:29:33 GMT */
+	int y;
+	struct tm tm;
+	memset(&tm, 0, sizeof(struct tm));
+
+	date = cast_uchar strchr(cast_const_char date, ' ');
+	if (!date) return (time_t)-1;
+	date++;
+	if (*date >= '0' && *date <= '9') {
+			/* Sun, 06 Nov 1994 08:49:37 GMT */
+			/* Sunday, 06-Nov-94 08:49:37 GMT */
+		y = 0;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+		tm.tm_mday = (date[0] - '0') * 10 + date[1] - '0';
+		date += 2;
+		if (*date != ' ' && *date != '-') return (time_t)-1;
+		date += 1;
+		for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++)
+			if (!casecmp(date, months[tm.tm_mon], 3)) goto f1;
+		return (time_t)-1;
+		f1:
+		date += 3;
+		if (*date != '-' && *date != ' ') return (time_t)-1;
+		date++;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+		if (date[2] == ' ') {
+				/* Sunday, 06-Nov-94 08:49:37 GMT */
+			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+			if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+			tm.tm_year = (date[0] >= '7' ? 1900 : 2000) + (date[0] - '0') * 10 + date[1] - '0' - 1900;
+			date += 2;
+		} else if (date[2] >= '0' && date[2] <= '9' && date[3] >= '0' && date[3] <= '9') {
+				/* Sun, 06 Nov 1994 08:49:37 GMT */
+			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+			if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+			if (date[2] < '0' || date[2] > '9') return (time_t)-1;
+			if (date[3] < '0' || date[3] > '9') return (time_t)-1;
+			tm.tm_year = (date[0] - '0') * 1000 + (date[1] - '0') * 100 + (date[2] - '0') * 10 + date[3] - '0' - 1900;
+			date += 4;
+		} else return (time_t)-1;
+		if (*date != ' ') return (time_t)-1;
+		date++;
+	} else {
+			/* Sun Nov  6 08:49:37 1994 */
+		y = 1;
+		for (tm.tm_mon = 0; tm.tm_mon < 12; tm.tm_mon++)
+			if (!casecmp(date, months[tm.tm_mon], 3)) goto f2;
+		return (time_t)-1;
+		f2:
+		date += 3;
+		while (*date == ' ') date++;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		tm.tm_mday = date[0] - '0';
+		date++;
+		if (*date != ' ') {
+			if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+			tm.tm_mday = tm.tm_mday * 10 + date[0] - '0';
+			date++;
+		}
+		if (*date != ' ') return (time_t)-1;
+		date++;
+	}
+
+	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+	tm.tm_hour = (date[0] - '0') * 10 + date[1] - '0';
+	date += 2;
+	if (*date != ':') return (time_t)-1;
+	date++;
+	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+	tm.tm_min = (date[0] - '0') * 10 + date[1] - '0';
+	date += 2;
+	if (*date != ':') return (time_t)-1;
+	date++;
+	if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+	if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+	tm.tm_sec = (date[0] - '0') * 10 + date[1] - '0';
+	date += 2;
+	if (y) {
+		if (*date != ' ') return (time_t)-1;
+		date++;
+		if (date[0] < '0' || date[0] > '9') return (time_t)-1;
+		if (date[1] < '0' || date[1] > '9') return (time_t)-1;
+		if (date[2] < '0' || date[2] > '9') return (time_t)-1;
+		if (date[3] < '0' || date[3] > '9') return (time_t)-1;
+		tm.tm_year = (date[0] - '0') * 1000 + (date[1] - '0') * 100 + (date[2] - '0') * 10 + date[3] - '0' - 1900;
+		date += 4;
+	}
+	if (*date != ' ' && *date) return (time_t)-1;
+
+	/*debug("%02d:%02d:%02d", tm.tm_hour, tm.tm_min, tm.tm_sec);*/
+#if defined(HAVE_TIMEGM) && !defined(HAIKU)
+	t = timegm(&tm);
+	if (t == (time_t)-1)
+		return (time_t)-1;
+	offset = 0;
+#else
+	t = mktime(&tm);
+	if (t == (time_t)-1)
+		return (time_t)-1;
+	memset(&tm, 0, sizeof(struct tm));
+	tm.tm_year = 80;
+	tm.tm_mday = 1;
+	offset = 315532800 - mktime(&tm);
+#endif
+	return (time_t)((uttime)t + (uttime)offset);
+#else
+	return (time_t)-1;
+#endif
+}
+
+unsigned char *print_http_date(time_t date)
+{
+	static unsigned char str[64];
+#ifdef HAVE_GMTIME
+	struct tm *tm = gmtime(&date);
+	snprintf(cast_char str, sizeof str, "%s %s %2d %02d:%02d:%02d %04d", days[tm->tm_wday], months[tm->tm_mon], tm->tm_mday, tm->tm_hour, tm->tm_min, tm->tm_sec, tm->tm_year + 1900);
+#else
+	snprintf(cast_char str, sizeof str, "%.0f", (double)date);
+#endif
+	return str;
+}
+
 static struct {
 	char *name;
 	int bugs;
@@ -149,6 +307,7 @@ static int check_http_server_bugs(unsigned char *url, struct http_connection_inf
 
 static void http_end_request(struct connection *c, int notrunc, int nokeepalive, int state)
 {
+	struct http_connection_info *info = c->info;
 	if (state == S__OK) {
 		if (c->cache) {
 			if (!notrunc) truncate_entry(c->cache, c->from, 1);
@@ -157,8 +316,8 @@ static void http_end_request(struct connection *c, int notrunc, int nokeepalive,
 	}
 	setcstate(c, state);
 	if (c->info &&
-	    !((struct http_connection_info *)c->info)->close &&
-	    !((struct http_connection_info *)c->info)->send_close &&
+	    !info->close &&
+	    !info->send_close &&
 	    !nokeepalive) {
 		add_keepalive_socket(c, HTTP_KEEPALIVE_TIMEOUT, 0);
 	} else {
@@ -221,14 +380,13 @@ static void http_send_header(struct connection *c)
 
 	proxy = is_proxy_url(c->url);
 	host = remove_proxy_prefix(c->url);
-	set_connection_timeout(c);
+	set_connection_timeout_keepal(c);
 	info = mem_calloc(sizeof(struct http_connection_info));
 	c->info = info;
 #ifdef HAVE_SSL
 	info->https_forward = !c->ssl && proxy && host && !casecmp(host, cast_uchar "https://", 8);
 	if (c->ssl) proxy = 0;
 #endif
-	info->send_close = info->https_forward || http10 || (post && http_options.bug_post_no_keepalive);
 	hdr = init_str();
 	if (!host) {
 		http_bad_url:
@@ -246,6 +404,7 @@ static void http_send_header(struct connection *c)
 		post = cast_uchar strchr(cast_const_char host, POST_CHAR);
 		if (post) post++;
 	}
+	info->send_close = info->https_forward || http10;
 #ifdef HAVE_SSL
 	if (info->https_forward) {
 		add_to_str(&hdr, &l, cast_uchar "CONNECT ");
@@ -268,7 +427,7 @@ static void http_send_header(struct connection *c)
 		c->unrestartable = 2;
 	}
 	if (!proxy) {
-		add_to_str(&hdr, &l, cast_uchar "/");
+		add_chr_to_str(&hdr, &l, '/');
 		u = get_url_data(host);
 	} else {
 		u = host;
@@ -309,13 +468,19 @@ static void http_send_header(struct connection *c)
 		add_to_str(&hdr, &l, h);
 		mem_free(h);
 		if ((h = get_port_str(host))) {
-			add_chr_to_str(&hdr, &l, ':');
-			add_to_str(&hdr, &l, h);
+			unsigned char *default_port = cast_uchar "80";
+#ifdef HAVE_SSL
+			if (c->ssl) default_port = cast_uchar "443";
+#endif
+			if (strcmp(cast_const_char h, cast_const_char default_port)) {
+				add_chr_to_str(&hdr, &l, ':');
+				add_to_str(&hdr, &l, h);
+			}
 			mem_free(h);
 		}
 		add_to_str(&hdr, &l, cast_uchar "\r\n");
 	}
-	add_user_agent(&hdr, &l);
+	add_user_agent(&hdr, &l, info->https_forward ? NULL : host);
 	if (proxy) add_proxy_auth_string(&hdr, &l, c->url);
 	if (!info->https_forward) {
 		test_restart(c);
@@ -374,12 +539,30 @@ static void test_restart(struct connection *c)
 	}
 }
 
-static void add_user_agent(unsigned char **hdr, int *l)
+static void add_user_agent(unsigned char **hdr, int *l, unsigned char *url)
 {
 	add_to_str(hdr, l, cast_uchar "User-Agent: ");
 	if (SCRUB_HEADERS) {
-		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 6.1; rv:52.0) Gecko/20100101 Firefox/52.0\r\n");
+		add_to_str(hdr, l, cast_uchar "Mozilla/5.0 (Windows NT 10.0; rv:78.0) Gecko/20100101 Firefox/78.0\r\n");
 	} else if (!(*http_options.header.fake_useragent)) {
+
+		/*
+		 * Google started to return css-styled page for searches.
+		 * It returns non-css page if the user agent begins with Lynx.
+		 */
+#if 0
+		if (url &&
+			(casestrstr(url, cast_uchar "/www.google.") ||
+			 casestrstr(url, cast_uchar "/google.")) &&
+			strstr(cast_const_char url, "/search?") &&
+			(strstr(cast_const_char url, "?q=") ||
+			 strstr(cast_const_char url, "&q=")) &&
+			!strstr(cast_const_char url, "?tbm=isch") &&
+			!strstr(cast_const_char url, "&tbm=isch")
+			)
+				add_to_str(hdr, l, cast_uchar("Lynx/"));
+#endif
+
 		add_to_str(hdr, l, cast_uchar("Links (" VERSION_STRING "; "));
 		add_to_str(hdr, l, system_name);
 		add_to_str(hdr, l, cast_uchar "; ");
@@ -484,7 +667,7 @@ static void add_accept_language(unsigned char **hdr, int *l, struct http_connect
 			int la;
 			la = *l;
 			add_to_str(hdr, l, get_text_translation(TEXT_(T__ACCEPT_LANGUAGE), NULL));
-			add_to_str(hdr, l, cast_uchar ",");
+			add_chr_to_str(hdr, l, ',');
 			if (!strstr(cast_const_char(*hdr + la), "en,") &&
 			    !strstr(cast_const_char(*hdr + la), "en;")) add_to_str(hdr, l, cast_uchar "en;q=0.2,");
 			add_to_str(hdr, l, cast_uchar "*;q=0.1\r\n");
@@ -535,6 +718,12 @@ static void add_accept_encoding(unsigned char **hdr, int *l, unsigned char *url,
 			add_to_str(hdr, l, cast_uchar "br");
 		}
 #endif
+#if defined(HAVE_ZSTD)
+		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
+			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
+			add_to_str(hdr, l, cast_uchar "zstd");
+		}
+#endif
 #if defined(HAVE_BZIP2)
 		if (!SCRUB_HEADERS && !(info->bl_flags & BL_NO_BZIP2)) {
 			if (*l != l1) add_to_str(hdr, l, cast_uchar ", ");
@@ -576,7 +765,7 @@ static void add_accept_charset(unsigned char **hdr, int *l, struct http_connecti
 		int aclen = 0;
 		ac = init_str();
 		for (i = 0; (cs = get_cp_mime_name(i)); i++) {
-			if (aclen) add_to_str(&ac, &aclen, cast_uchar ",");
+			if (aclen) add_chr_to_str(&ac, &aclen, ',');
 			else add_to_str(&ac, &aclen, cast_uchar "Accept-Charset: ");
 			add_to_str(&ac, &aclen, cs);
 		}
@@ -773,7 +962,7 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 		if (info->length >= 0) info->length -= l;
 		c->from += l;
 		kill_buffer_data(rb, l);
-		if (!info->length && !rb->close) {
+		if (!info->length) {
 			http_end_request(c, 0, 0, S__OK);
 			return;
 		}
@@ -797,10 +986,10 @@ static void read_http_data(struct connection *c, struct read_buffer *rb)
 		} else if (info->chunk_remaining == -1) {
 			int l;
 			if ((l = is_line_in_buffer(rb))) {
-				unsigned char *de;
+				char *end;
 				long n = 0;	/* warning, go away */
-				if (l != -1) n = strtol(cast_const_char rb->data, (char **)(void *)&de, 16);
-				if (l == -1 || n < 0 || n >= MAXINT || de == rb->data) {
+				if (l != -1) n = strtol(cast_const_char rb->data, &end, 16);
+				if (l == -1 || n < 0 || n >= MAXINT || cast_uchar end == rb->data) {
 					setcstate(c, S_HTTP_ERROR);
 					abort_connection(c);
 					return;
@@ -916,8 +1105,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		return;
 	}
 	if (a != -2) {
-		head = mem_alloc(a + 1);
-		memcpy(head, rb->data, a); head[a] = 0;
+		head = memacpy(rb->data, a);
 		kill_buffer_data(rb, a);
 	} else {
 		head = stracpy(cast_uchar "HTTP/0.9 200 OK\r\nContent-Type: text/html\r\n\r\n");
@@ -965,8 +1153,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		unsigned char *cookie;
 		unsigned char *ch = head;
 		while ((cookie = parse_http_header(ch, cast_uchar "Set-Cookie", &ch))) {
-			unsigned char *host = remove_proxy_prefix(c->url);
-			set_cookie(NULL, host, cookie);
+			set_cookie(host, cookie);
 			mem_free(cookie);
 		}
 	}
@@ -1027,7 +1214,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	e->head = head;
 	if ((d = parse_http_header(head, cast_uchar "Expires", NULL))) {
 		time_t t = parse_http_date(d);
-		if (t && e->expire_time != 1) e->expire_time = t;
+		if (t != (time_t)-1 && e->expire_time != 1) e->expire_time = t;
 		mem_free(d);
 	}
 	if ((d = parse_http_header(head, cast_uchar "Pragma", NULL))) {
@@ -1044,9 +1231,8 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 			}
 			if (!casecmp(f, cast_uchar "max-age=", 8)) {
 				if (e->expire_time != 1) {
-					errno = 0;
-					EINTRLOOPX(e->expire_time, time(NULL), (time_t)-1);
-					e->expire_time += atoi(cast_const_char(f + 8));
+					e->expire_time = get_absolute_seconds();
+					e->expire_time = (time_t)((uttime)e->expire_time + (uttime)atoi(cast_const_char(f + 8)));
 				}
 			}
 			while (*f && *f != ',') f++;
@@ -1057,11 +1243,13 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 	if (c->ssl) {
 		if (e->ssl_info) mem_free(e->ssl_info);
 		e->ssl_info = get_cipher_string(c->ssl);
+		if (e->ssl_authority) mem_free(e->ssl_authority);
+		e->ssl_authority = stracpy(c->ssl->ca);
 	}
 #endif
 	if (e->redirect) mem_free(e->redirect), e->redirect = NULL;
 	if ((h == 302 || h == 303 || h == 307 || h == 511) && !e->expire_time) e->expire_time = 1;
-	if (h == 301 || h == 302 || h == 303 || h == 307) {
+	if (h == 301 || h == 302 || h == 303 || h == 307 || h == 308) {
 		if ((d = parse_http_header(e->head, cast_uchar "Location", NULL))) {
 			unsigned char *user, *ins;
 			unsigned char *newuser, *newpassword;
@@ -1083,7 +1271,11 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 			}
 			if (e->redirect) mem_free(e->redirect);
 			e->redirect = d;
-			e->redirect_get = h == 303;
+			if (h == 307 || h == 308) {
+				unsigned char *p;
+				if ((p = cast_uchar strchr(cast_const_char host, POST_CHAR)))
+					add_to_strn(&e->redirect, p);
+			}
 		}
 	}
 	if (!e->expire_time && strchr(cast_const_char c->url, POST_CHAR)) e->expire_time = 1;
@@ -1119,7 +1311,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 		unsigned char *ep;
 		my_strtoll_t l = my_strtoll(d, &ep);
 		if (!*ep && l >= 0 && (off_t)l >= 0 && (off_t)l == l) {
-			if (!info->close || version >= 11) info->length = l;
+			if (!info->close || version >= 11 || h / 100 == 3) info->length = l;
 			if (c->from + l >= 0) c->est_length = c->from + l;
 		}
 		mem_free(d);
@@ -1183,7 +1375,7 @@ static void http_got_header(struct connection *c, struct read_buffer *rb)
 static void http_get_header(struct connection *c)
 {
 	struct read_buffer *rb;
-	set_connection_timeout(c);
+	set_connection_timeout_keepal(c);
 	if (!(rb = alloc_read_buffer(c))) return;
 	rb->close = 1;
 	read_from_socket(c, c->sock1, rb, http_got_header);
